"""titiler.core.algorithm Math."""

import numpy
from rio_tiler.models import ImageData

from titiler.core.algorithm.base import BaseAlgorithm

__all__ = ["_Min", "_Max", "_Median", "_Mean", "_Std", "_Var", "_Sum"]


class _Min(BaseAlgorithm):
    """Return Min values along the `bands` axis."""

    title: str = "Min"
    description: str = "Return Min values along the `bands` axis."

    output_nbands: int = 1

    def __call__(self, img: ImageData) -> ImageData:
        """Return Min."""
        return ImageData(
            numpy.ma.min(img.array, axis=0, keepdims=True),
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
            band_descriptions=["min"],
            metadata=img.metadata,
            cutline_mask=img.cutline_mask,
        )


class _Max(BaseAlgorithm):
    """Return Max values along the `bands` axis."""

    title: str = "Max"
    description: str = "Return Max values along the `bands` axis."

    output_nbands: int = 1

    def __call__(self, img: ImageData) -> ImageData:
        """Return Max."""
        return ImageData(
            numpy.ma.max(img.array, axis=0, keepdims=True),
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
            band_descriptions=["max"],
            metadata=img.metadata,
            cutline_mask=img.cutline_mask,
        )


class _Median(BaseAlgorithm):
    """Return Median values along the `bands` axis."""

    title: str = "Median"
    description: str = "Return Median values along the `bands` axis."

    output_nbands: int = 1
    output_dtype: str = "float64"

    def __call__(self, img: ImageData) -> ImageData:
        """Return Median."""
        return ImageData(
            numpy.ma.median(img.array, axis=0, keepdims=True),
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
            band_descriptions=["median"],
            metadata=img.metadata,
            cutline_mask=img.cutline_mask,
        )


class _Mean(BaseAlgorithm):
    """Return Mean values along the `bands` axis."""

    title: str = "Mean"
    description: str = "Return Mean values."

    output_nbands: int = 1
    output_dtype: str = "float64"

    def __call__(self, img: ImageData) -> ImageData:
        """Return Mean."""
        return ImageData(
            numpy.ma.mean(img.array, axis=0, keepdims=True),
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
            band_descriptions=["mean"],
            metadata=img.metadata,
            cutline_mask=img.cutline_mask,
        )


class _Std(BaseAlgorithm):
    """Return the standard deviation along the `bands` axis."""

    title: str = "Standard deviation"
    description: str = "Return the Standard Deviation along the `bands` axis."

    output_nbands: int = 1
    output_dtype: str = "float64"

    def __call__(self, img: ImageData) -> ImageData:
        """Return Stddev."""
        return ImageData(
            numpy.ma.std(img.array, axis=0, keepdims=True, ddof=1),
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
            band_descriptions=["std"],
            metadata=img.metadata,
            cutline_mask=img.cutline_mask,
        )


class _Var(BaseAlgorithm):
    """Return Variance values along the `bands` axis."""

    title: str = "Variance"
    description: str = "Return Variance along the `bands` axis."

    output_nbands: int = 1
    output_dtype: str = "float64"

    def __call__(self, img: ImageData) -> ImageData:
        """Return Variance."""
        return ImageData(
            numpy.ma.var(img.array, axis=0, keepdims=True, ddof=1),
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
            band_descriptions=["var"],
            metadata=img.metadata,
            cutline_mask=img.cutline_mask,
        )


class _Sum(BaseAlgorithm):
    """Return Sum values along the `bands` axis."""

    title: str = "Sum"
    description: str = "Return Sum values along the `bands` axis."

    output_nbands: int = 1

    def __call__(self, img: ImageData) -> ImageData:
        """Return Min."""
        return ImageData(
            numpy.ma.sum(img.array, axis=0, keepdims=True),
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
            band_descriptions=["sum"],
            metadata=img.metadata,
            cutline_mask=img.cutline_mask,
        )



"""titiler.core.algorithm Ops."""

from collections.abc import Sequence

import numpy
from rio_tiler.models import ImageData

from titiler.core.algorithm.base import BaseAlgorithm

__all__ = ["CastToInt", "Ceil", "Floor"]


class CastToInt(BaseAlgorithm):
    """Cast data to Integer."""

    title: str = "Cast data to Integer"
    description: str = "Cast data to Integer."

    # metadata
    output_dtype: str = "uint8"
    output_min: Sequence[int] = [0]
    output_max: Sequence[int] = [255]

    def __call__(self, img: ImageData) -> ImageData:
        """Cast Data."""
        return ImageData(
            img.array.astype("uint8"),
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
            band_descriptions=img.band_descriptions,
            metadata=img.metadata,
            cutline_mask=img.cutline_mask,
        )


class Ceil(BaseAlgorithm):
    """Round data to the smallest integer."""

    title: str = "Round data to the smallest integer"
    description: str = "Round data to the smallest integer."

    # metadata
    output_dtype: str = "uint8"
    output_min: Sequence[int] = [0]
    output_max: Sequence[int] = [255]

    def __call__(self, img: ImageData) -> ImageData:
        """Cast Data."""
        return ImageData(
            numpy.ceil(img.array).astype("uint8"),
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
            band_descriptions=img.band_descriptions,
            metadata=img.metadata,
            cutline_mask=img.cutline_mask,
        )


class Floor(BaseAlgorithm):
    """Round data to the largest integer."""

    title: str = "Round data to the largest integer"
    description: str = "Round data to the largest integer."

    # metadata
    output_dtype: str = "uint8"
    output_min: Sequence[int] = [0]
    output_max: Sequence[int] = [255]

    def __call__(self, img: ImageData) -> ImageData:
        """Cast Data."""
        return ImageData(
            numpy.floor(img.array).astype("uint8"),
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
            band_descriptions=img.band_descriptions,
            metadata=img.metadata,
            cutline_mask=img.cutline_mask,
        )



"""OpenTelemetry instrumentation for titiler.core."""

from __future__ import annotations

import functools
import inspect
from collections.abc import Callable
from contextlib import contextmanager
from typing import Any, Iterator, ParamSpec, TypeVar

from titiler.core import __version__

try:
    from opentelemetry import trace
    from opentelemetry.trace import Span, Status, StatusCode

    tracer = trace.get_tracer("titiler.core", __version__)
except ImportError:
    trace = None  # type: ignore
    Span = None  # type: ignore
    Status = None  # type: ignore
    StatusCode = None  # type: ignore
    tracer = None  # type: ignore

P = ParamSpec("P")
R = TypeVar("R")


def add_span_attributes(attributes: dict[str, Any]) -> None:
    """Adds attributes to the current active span."""
    if not tracer:
        return
    span = trace.get_current_span()
    if span and span.is_recording():
        span.set_attributes(attributes)


def flatten_dict(d: dict, parent_key: str = "", sep: str = ".") -> dict:
    """Flattens a nested dictionary for adding span attributes."""
    items = {}
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.update(flatten_dict(v, new_key, sep=sep))
        else:
            if not isinstance(v, (str, bool, int, float)):
                v = str(v)
            items[new_key] = v
    return items


class SpanWrapper:
    """A wrapper class to safely handle an optional OpenTelemetry Span."""

    def __init__(self, span: Span | None):
        """Set the span"""
        self._span = span

    def set_attributes(self, attributes: dict[str, Any]) -> None:
        """Safely set attributes on the wrapped span if it exists."""
        if self._span:
            self._span.set_attributes(attributes)

    def record_exception(self, exception: Exception) -> None:
        """Safely record an exception on the wrapped span if it exists."""
        if self._span:
            self._span.record_exception(exception)


@contextmanager
def operation_tracer(
    operation_name: str,
    attributes: dict[str, Any] | None = None,
) -> Iterator[SpanWrapper]:
    """Context manager for creating granular child spans."""
    if not tracer:
        yield SpanWrapper(None)
        return

    with tracer.start_as_current_span(operation_name) as span:
        wrapped_span = SpanWrapper(span)
        if attributes:
            wrapped_span.set_attributes(attributes)
        try:
            yield wrapped_span
            span.set_status(Status(StatusCode.OK))
        except Exception as e:
            span.set_status(Status(StatusCode.ERROR, str(e)))
            raise


def _get_span_name(op_name: str, factory_instance: Any) -> str:
    """Determine the span name using the factory class name."""
    if not factory_instance:
        return op_name
    class_name = factory_instance.__class__.__name__
    return f"{class_name}.{op_name}"


def factory_trace(
    _func: Callable[P, Any] | None = None,
    *,
    factory_instance: Any | None = None,
) -> Any:
    """A decorator for Factory methods that automatically handles tracing for factory methods"""

    def decorator(func: Callable[P, Any]) -> Callable[P, Any]:
        if not tracer:
            return func

        op_name = func.__name__

        attributes = {}
        if factory_instance:
            if hasattr(factory_instance, "reader"):
                attributes["reader"] = str(factory_instance.reader)
            if hasattr(factory_instance, "backend"):
                attributes["backend"] = str(factory_instance.backend)

        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
                span_name = _get_span_name(op_name, factory_instance)
                with operation_tracer(span_name, attributes=attributes):
                    return await func(*args, **kwargs)

            return async_wrapper
        else:

            @functools.wraps(func)
            def sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
                span_name = _get_span_name(op_name, factory_instance)
                with operation_tracer(span_name, attributes=attributes):
                    return func(*args, **kwargs)

            return sync_wrapper

    return decorator if _func is None else decorator(_func)


factory_trace.decorator_enabled = bool(tracer)  # type: ignore [attr-defined]



"""titiler.core"""

__version__ = "2.0.0a1"

from . import dependencies, errors, factory, routing  # noqa
from .factory import (  # noqa
    AlgorithmFactory,
    BaseFactory,
    ColorMapFactory,
    MultiBaseTilerFactory,
    TilerFactory,
    TMSFactory,
)



"""Common dependency."""

import json
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import Annotated, Literal

import numpy
from fastapi import HTTPException, Query
from pydantic import BeforeValidator, Field
from rasterio.crs import CRS
from rio_tiler.colormap import ColorMaps
from rio_tiler.colormap import cmap as default_cmap
from rio_tiler.colormap import parse_color
from rio_tiler.types import RIOResampling, WarpResampling
from starlette.requests import Request

from titiler.core.resources.enums import ImageType, MediaType
from titiler.core.utils import accept_media_type
from titiler.core.validation import (
    separated_parseable_floats_regex,
    validate_bbox,
    validate_color_formula,
    validate_crs,
    validate_rescale,
)


def create_colormap_dependency(cmap: ColorMaps) -> Callable:
    """Create Colormap Dependency."""

    def deps(
        colormap_name: Annotated[  # type: ignore
            Literal[tuple(cmap.list())],
            Query(description="Colormap name"),
        ] = None,
        colormap: Annotated[
            str | None, Query(description="JSON encoded custom Colormap")
        ] = None,
    ):
        if colormap_name:
            return cmap.get(colormap_name)

        if colormap:
            try:
                c = json.loads(
                    colormap,
                    object_hook=lambda x: {
                        int(k): parse_color(v) for k, v in x.items()
                    },
                )

                # Make sure to match colormap type
                if isinstance(c, Sequence):
                    c = [(tuple(inter), parse_color(v)) for (inter, v) in c]

                return c
            except json.JSONDecodeError as e:
                raise HTTPException(
                    status_code=400, detail="Could not parse the colormap value."
                ) from e

        return None

    return deps


ColorMapParams = create_colormap_dependency(default_cmap)


def DatasetPathParams(url: Annotated[str, Query(description="Dataset URL")]) -> str:
    """Create dataset path from args"""
    return url


@dataclass
class DefaultDependency:
    """Dataclass with dict unpacking"""

    def as_dict(self, exclude_none: bool = True) -> dict:
        """Transform dataclass to dict."""
        if exclude_none:
            return {k: v for k, v in self.__dict__.items() if v is not None}

        return dict(self.__dict__.items())


# Dependencies for simple BaseReader (e.g COGReader)
@dataclass
class BidxParams(DefaultDependency):
    """Band Indexes parameters."""

    indexes: Annotated[
        list[int] | None,
        Query(
            title="Band indexes",
            alias="bidx",
            description="Dataset band indexes",
            openapi_examples={
                "user-provided": {"value": None},
                "one-band": {"value": [1]},
                "multi-bands": {"value": [1, 2, 3]},
            },
        ),
    ] = None


@dataclass
class ExpressionParams(DefaultDependency):
    """Expression parameters."""

    expression: Annotated[
        str | None,
        Query(
            title="Band Math expression",
            description="rio-tiler's band math expression",
            openapi_examples={
                "user-provided": {"value": None},
                "simple": {"description": "Simple band math.", "value": "b1/b2"},
                "multi-bands": {
                    "description": "Semicolon (;) delimited expressions (band1: b1/b2, band2: b2+b3).",
                    "value": "b1/b2;b2+b3",
                },
            },
        ),
    ] = None


@dataclass
class BidxExprParams(ExpressionParams, BidxParams):
    """Band Indexes and Expression parameters."""

    pass


# Dependencies for  MultiBaseReader (e.g STACReader)
@dataclass
class AssetsParams(DefaultDependency):
    """Assets parameters."""

    assets: Annotated[
        list[str],
        Query(
            title="Asset names",
            description="Asset's names.",
            openapi_examples={
                "user-provided": {"value": None},
                "one-asset": {
                    "description": "Return results for asset `data`.",
                    "value": ["data"],
                },
                "multi-assets": {
                    "description": "Return results for assets `data` and `cog`.",
                    "value": ["data", "cog"],
                },
            },
        ),
    ]


@dataclass
class AssetsExprParams(ExpressionParams, AssetsParams):
    """Assets and Expression parameters."""

    asset_as_band: Annotated[
        bool | None,
        Query(
            title="Consider asset as a 1 band dataset",
            description="Asset as Band",
        ),
    ] = None


@dataclass
class PreviewParams(DefaultDependency):
    """Common Preview parameters."""

    # NOTE: sizes dependency can either be a Query or a Path Parameter
    max_size: Annotated[int, Field(description="Maximum image size to read onto.")] = (
        1024
    )
    height: Annotated[int | None, Field(description="Force output image height.")] = (
        None
    )
    width: Annotated[int | None, Field(description="Force output image width.")] = None

    def __post_init__(self):
        """Post Init."""
        if self.width or self.height:
            self.max_size = None


@dataclass
class PartFeatureParams(DefaultDependency):
    """Common parameters for bbox and feature."""

    # NOTE: the part sizes dependency can either be a Query or a Path Parameter
    max_size: Annotated[
        int | None, Field(description="Maximum image size to read onto.")
    ] = None
    height: Annotated[int | None, Field(description="Force output image height.")] = (
        None
    )
    width: Annotated[int | None, Field(description="Force output image width.")] = None

    def __post_init__(self):
        """Post Init."""
        if self.width or self.height:
            self.max_size = None


@dataclass
class DatasetParams(DefaultDependency):
    """Low level WarpedVRT Optional parameters."""

    nodata: Annotated[
        Literal["nan", "inf", "-inf"] | int | float | None,
        Query(
            title="Nodata value",
            description="Overwrite internal Nodata value; nan or valid float values only.",
        ),
    ] = None
    unscale: Annotated[
        bool | None,
        Query(
            title="Apply internal Scale/Offset",
            description="Apply internal Scale/Offset. Defaults to `False`.",
        ),
    ] = None
    resampling_method: Annotated[
        RIOResampling | None,
        Query(
            alias="resampling",
            description="RasterIO resampling algorithm. Defaults to `nearest`.",
        ),
    ] = None
    reproject_method: Annotated[
        WarpResampling | None,
        Query(
            alias="reproject",
            description="WarpKernel resampling algorithm (only used when doing re-projection). Defaults to `nearest`.",
        ),
    ] = None

    def __post_init__(self):
        """Post Init."""
        if self.nodata is not None:
            self.nodata = numpy.nan if self.nodata == "nan" else float(self.nodata)

        if self.unscale is not None:
            self.unscale = bool(self.unscale)


RescaleType = list[tuple[float, float]]


@dataclass
class RenderingParams(DefaultDependency):
    """Image Rendering options."""

    rescale: Annotated[
        list[str] | None,
        BeforeValidator(validate_rescale),
        Query(
            title="Min/Max data Rescaling",
            description="comma (',') delimited Min,Max range. Can set multiple time for multiple bands.",
            examples=["0,2000", "0,1000", "0,10000"],  # band 1  # band 2  # band 3
        ),
    ] = None

    color_formula: Annotated[
        str | None,
        BeforeValidator(validate_color_formula),
        Query(
            title="Color Formula",
            description="rio-color formula (info: https://github.com/mapbox/rio-color)",
        ),
    ] = None

    def __post_init__(self) -> None:
        """Post Init."""
        if self.rescale:
            rescale_array = []
            for r in self.rescale:
                parsed = tuple(
                    map(
                        float,
                        r.split(","),
                    )
                )
                assert (
                    len(parsed) == 2
                ), f"Invalid rescale values: {self.rescale}, should be of form ['min,max', 'min,max'] or [[min,max], [min, max]]"
                rescale_array.append(parsed)

            self.rescale: RescaleType = rescale_array  # type: ignore


@dataclass
class ImageRenderingParams(RenderingParams):
    """Image Rendering options."""

    add_mask: Annotated[
        bool | None,
        Query(
            alias="return_mask",
            description="Add mask to the output data. Defaults to `True`",
        ),
    ] = None


@dataclass
class StatisticsParams(DefaultDependency):
    """Statistics options."""

    categorical: Annotated[
        bool | None,
        Query(
            description="Return statistics for categorical dataset. Defaults to `False`"
        ),
    ] = None
    categories: Annotated[
        list[float | int] | None,
        Query(
            alias="c",
            title="Pixels values for categories.",
            description="List of values for which to report counts.",
            examples=[1, 2, 3],
        ),
    ] = None
    percentiles: Annotated[
        list[int] | None,
        Query(
            alias="p",
            title="Percentile values",
            description="List of percentile values (default to [2, 98]).",
            examples=[2, 5, 95, 98],
        ),
    ] = None

    def __post_init__(self):
        """Set percentiles default."""
        if not self.percentiles:
            self.percentiles = [2, 98]


@dataclass
class HistogramParams(DefaultDependency):
    """Numpy Histogram options."""

    bins: Annotated[
        str | None,
        Query(
            alias="histogram_bins",
            title="Histogram bins.",
            description="""
Defines the number of equal-width bins in the given range (10, by default).

If bins is a sequence (comma `,` delimited values), it defines a monotonically increasing array of bin edges, including the rightmost edge, allowing for non-uniform bin widths.

link: https://numpy.org/doc/stable/reference/generated/numpy.histogram.html
            """,
            openapi_examples={
                "user-provided": {"value": None},
                "simple": {
                    "description": "Defines the number of equal-width bins",
                    "value": 8,
                },
                "array": {
                    "description": "Defines custom bin edges (comma `,` delimited values)",
                    "value": "0,100,200,300",
                },
            },
            pattern=r"^\d+(,\d+)*$",
        ),
    ] = None

    range: Annotated[
        str | None,
        Query(
            alias="histogram_range",
            title="Histogram range",
            description="""
Comma `,` delimited range of the bins.

The lower and upper range of the bins. If not provided, range is simply (a.min(), a.max()).

Values outside the range are ignored. The first element of the range must be less than or equal to the second.
range affects the automatic bin computation as well.

link: https://numpy.org/doc/stable/reference/generated/numpy.histogram.html
            """,
            openapi_examples={
                "user-provided": {"value": None},
                "array": {
                    "description": "Defines custom histogram range (comma `,` delimited values)",
                    "value": "0,1000",
                },
            },
            pattern=separated_parseable_floats_regex(count=2),
        ),
    ] = None

    def __post_init__(self):
        """Post Init."""
        if self.bins:
            bins = self.bins.split(",")
            if len(bins) == 1:
                self.bins = int(bins[0])  # type: ignore
            else:
                self.bins = list(map(float, bins))  # type: ignore
        else:
            self.bins = 10

        if self.range:
            parsed = list(map(float, self.range.split(",")))
            assert (
                len(parsed) == 2
            ), f"Invalid histogram_range values: {self.range}, should be of form 'min,max'"

            self.range = parsed  # type: ignore


def CoordCRSParams(
    crs: Annotated[
        str | None,
        BeforeValidator(validate_crs),
        Query(
            alias="coord_crs",
            description="Coordinate Reference System of the input coords. Default to `epsg:4326`.",
        ),
    ] = None,
) -> CRS | None:
    """Coordinate Reference System Coordinates Param."""
    if crs:
        return CRS.from_user_input(crs)

    return None


def DstCRSParams(
    crs: Annotated[
        str | None,
        BeforeValidator(validate_crs),
        Query(
            alias="dst_crs",
            description="Output Coordinate Reference System.",
        ),
    ] = None,
) -> CRS | None:
    """Coordinate Reference System Coordinates Param."""
    if crs:
        return CRS.from_user_input(crs)

    return None


def CRSParams(
    crs: Annotated[
        str | None,
        BeforeValidator(validate_crs),
        Query(
            description="Coordinate Reference System.",
        ),
    ] = None,
) -> CRS | None:
    """Coordinate Reference System Coordinates Param."""
    if crs:
        return CRS.from_user_input(crs)

    return None


def BufferParams(
    buffer: Annotated[
        float | None,
        Query(
            gt=0,
            title="Tile buffer.",
            description="Buffer on each side of the given tile. It must be a multiple of `0.5`. Output **tilesize** will be expanded to `tilesize + 2 * buffer` (e.g 0.5 = 257x257, 1.0 = 258x258).",
        ),
    ] = None,
) -> float | None:
    """Tile buffer Parameter."""
    return buffer


@dataclass
class TileParams(DefaultDependency):
    """Tile options."""

    buffer: Annotated[
        float | None,
        Query(
            gt=0,
            title="Tile buffer.",
            description="Buffer on each side of the given tile. It must be a multiple of `0.5`. Output **tilesize** will be expanded to `tilesize + 2 * buffer` (e.g 0.5 = 257x257, 1.0 = 258x258).",
        ),
    ] = None

    padding: Annotated[
        int | None,
        Query(
            gt=0,
            title="Tile padding.",
            description="Padding to apply to each tile edge. Helps reduce resampling artefacts along edges. Defaults to `0`.",
        ),
    ] = None


@dataclass
class OGCMapsParams(DefaultDependency):
    """OGC Maps options."""

    request: Request

    bbox: Annotated[
        str | None,
        BeforeValidator(validate_bbox),
        Query(
            description="Bounding box of the rendered map. The bounding box is provided as four or six coordinates.",
        ),
    ] = None

    crs: Annotated[
        str | None,
        BeforeValidator(validate_crs),
        Query(
            description="Reproject the output to the given crs.",
        ),
    ] = None

    bbox_crs: Annotated[
        str | None,
        BeforeValidator(validate_crs),
        Query(
            description="crs for the specified bbox.",
            alias="bbox-crs",
        ),
    ] = None

    height: Annotated[
        int | None,
        Query(
            description="Height of the map in pixels. If omitted and `width` is specified, defaults to the `height` maintaining a 1:1 aspect ratio. If both `width` and `height` are omitted, the server will select default dimensions.",
            gt=0,
        ),
    ] = None

    width: Annotated[
        int | None,
        Query(
            description="Width of the map in pixels. If omitted and `height` is specified, defaults to the `width` maintaining a 1:1 aspect ratio. If both `width` and `height` are omitted, the server will select default dimensions.",
            gt=0,
        ),
    ] = None

    f: Annotated[
        ImageType | None,
        Query(description="The format of the map response (e.g. png)."),
    ] = None

    max_size: int | None = field(init=False, default=None)

    format: ImageType | None = field(init=False, default=ImageType.png)

    def __post_init__(self):  # noqa: C901
        """Parse and validate."""
        if self.crs:
            self.crs = CRS.from_user_input(self.crs)  # type: ignore

        if self.bbox_crs:
            self.bbox_crs = CRS.from_user_input(self.bbox_crs)  # type: ignore

        if not self.height and not self.width:
            self.max_size = 1024

        if self.bbox:
            bounds = list(map(float, self.bbox.split(",")))
            if len(bounds) == 6:
                bounds = [bounds[0], bounds[1], bounds[3], bounds[4]]

            self.bbox = bounds  # type: ignore

        if self.f:
            self.format = ImageType[self.f]

        else:
            if media := accept_media_type(
                self.request.headers.get("accept", ""),
                [MediaType[e] for e in ImageType],
            ):
                self.format = ImageType[media.name]



"""Dependency validations."""

import re
from json import JSONDecodeError, loads

from color_operations import parse_operations as parse_color_formula
from rasterio.crs import CRS


def validate_rescale(rescale_strs: list[str]) -> list[str]:
    """
    Verify that rescale input matches an accepted pattern.
    :param rescale_strs: Caller-provided rescale values.
    :type rescale_strs: list[str]
    :return: Caller-provided rescale values if validated, otherwise an exception is raised.
    :rtype: list[str]
    """
    validated_rescales: list[str] = []
    for rescale_str in rescale_strs:
        error_text = "invalid rescale format"
        rescale_parts = [
            re.sub(r"[\[|\]]", "", part) for part in rescale_str.split(",")
        ]
        if len(rescale_parts) == 2:
            try:
                # Regex validation adds risk, given the different string
                # formats that can be parsed to float, so simply attempt
                # to parse.
                min = float(rescale_parts[0])
                max = float(rescale_parts[1])
            except ValueError as e:
                error_text = f"{error_text}: {e}"
            else:
                validated_rescales.append(f"{min},{max}")
                continue
        raise ValueError(error_text)
    return validated_rescales


def validate_crs(crs_str: str | None) -> str | None:
    """
    Verify that crs input matches an accepted format.
    :param crs_str: Caller-provided crs value.
    :type crs_str: str | None
    :return: Caller-provided crs value if accepted, otherwise an exception is raised.
    :rtype: str | None
    """
    if crs_str is None:
        return None
    if crs_str.startswith("[") and crs_str.endswith(
        "]"
    ):  # this block lifted from OGCMapsParams
        crs_str = crs_str[1:-1]
    try:
        # CRS does not provide a static `is_valid` check, and "many different kinds" of formats
        # are supported, making regex impractical.
        # (https://rasterio.readthedocs.io/en/latest/api/rasterio.crs.html#rasterio.crs.CRS.from_user_input)
        CRS.from_user_input(crs_str)
    except ValueError as e:
        raise ValueError("invalid CRS format") from e
    else:
        return crs_str


def validate_json(json_str: str | None) -> str | None:
    """
    Verify that json input can be parsed.
    :param json_str: Caller-provided crs value.
    :type json_str: str | None
    :return: Caller-provided json_str value if valid, otherwise an exception is raised.
    :rtype: str | None
    """
    if json_str is None:
        return None
    try:
        loads(json_str)
    except JSONDecodeError as e:
        raise ValueError("invalid JSON content") from e
    else:
        return json_str


def validate_color_formula(color_formula_str: str | None) -> str | None:
    """
    Verify that color formula can be parsed.
    :param color_formula_str: Caller-provided color formula value.
    :type color_formula_str: str | None
    :return: Caller-provided color formula value if valid, otherwise an exception is raised.
    :rtype: str | None
    """
    if color_formula_str is None:
        return None
    try:
        parse_color_formula(color_formula_str)
    except Exception as e:
        raise ValueError("invalid color formula") from e
    else:
        return color_formula_str


def validate_bbox(bbox_str: str | None) -> str | None:
    """
    Verify that bbox can be parsed.
    :param bbox_str: Caller-provided bbox value.
    :type bbox_str: str | None
    :return: Caller-provided bbox_str value if valid, otherwise an exception is raised.
    :rtype: str | None
    """
    if bbox_str is None:
        return None
    if re.match(separated_parseable_floats_regex(count=4), bbox_str) or re.match(
        separated_parseable_floats_regex(count=6), bbox_str
    ):
        return bbox_str
    raise ValueError("invalid bbox content")


def separated_parseable_floats_regex(count: int, separator: str = ",") -> str:
    """
    Generate a regular expression for n separated strings representing floats.
    :param count: Number of separated float strings to match
    :type count: int
    :param separator: Character used to separate float strings.
    :type separator: str
    :return: Generated regular expression.
    :rtype: str
    """
    parseable_float_regex = r"\s*(-)?\d+((\.\d+)(e\d+)?)?\s*"
    return "^{}$".format(
        re.escape(separator).join([parseable_float_regex for _ in range(count)])
    )



"""TiTiler response models."""

from geojson_pydantic.features import Feature, FeatureCollection
from geojson_pydantic.geometries import Geometry, MultiPolygon, Polygon
from pydantic import BaseModel
from rio_tiler.models import BandStatistics, Info

from titiler.core.models.common import Link


class Point(BaseModel):
    """
    Point model.

    response model for `/point` endpoints

    """

    coordinates: list[float]
    values: list[float | None]
    band_names: list[str]
    band_descriptions: list[str] | None = None


InfoGeoJSON = Feature[Polygon | MultiPolygon, Info]
Statistics = dict[str, BandStatistics]


class StatisticsInGeoJSON(BaseModel):
    """Statistics model in geojson response."""

    statistics: Statistics

    model_config = {"extra": "allow"}


StatisticsGeoJSON = (
    FeatureCollection[Feature[Geometry, StatisticsInGeoJSON]]
    | Feature[Geometry, StatisticsInGeoJSON]
)

# MultiBase Models
MultiBaseInfo = dict[str, Info]
MultiBaseInfoGeoJSON = Feature[Polygon | MultiPolygon, MultiBaseInfo]

MultiBaseStatistics = dict[str, Statistics]
MultiBaseStatisticsGeoJSON = StatisticsGeoJSON


class ColorMapRef(BaseModel):
    """ColorMapRef model."""

    id: str
    title: str | None = None
    links: list[Link]


class ColorMapList(BaseModel):
    """Model for colormap list."""

    colormaps: list[ColorMapRef]



"""Custom routing classes."""

from typing import TypedDict

from fastapi import params
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from starlette.routing import BaseRoute, Match


class EndpointScope(TypedDict, total=False):
    """Define endpoint."""

    # More strict version of Starlette's Scope
    # https://github.com/encode/starlette/blob/6af5c515e0a896cbf3f86ee043b88f6c24200bcf/starlette/types.py#L3
    path: str
    method: str
    type: str | None  # http or websocket


def add_route_dependencies(
    routes: list[BaseRoute],
    *,
    scopes: list[EndpointScope],
    dependencies=list[params.Depends],
):
    """Add dependencies to routes.

    Allows a developer to add dependencies to a route after the route has been defined.

    """
    for route in routes:
        for scope in scopes:
            match, _ = route.matches({"type": "http", **scope})  # type: ignore
            if match != Match.FULL:
                continue

            # Mimicking how APIRoute handles dependencies:
            # https://github.com/tiangolo/fastapi/blob/1760da0efa55585c19835d81afa8ca386036c325/fastapi/routing.py#L408-L412
            for depends in dependencies[::-1]:
                route.dependant.dependencies.insert(  # type: ignore
                    0,
                    get_parameterless_sub_dependant(
                        depends=depends,
                        path=route.path_format,  # type: ignore
                    ),
                )

            # Register dependencies directly on route so that they aren't ignored if
            # the routes are later associated with an app (e.g. app.include_router(router))
            # https://github.com/tiangolo/fastapi/blob/58ab733f19846b4875c5b79bfb1f4d1cb7f4823f/fastapi/applications.py#L337-L360
            # https://github.com/tiangolo/fastapi/blob/58ab733f19846b4875c5b79bfb1f4d1cb7f4823f/fastapi/routing.py#L677-L678
            route.dependencies.extend(dependencies)  # type: ignore



"""titiler.core.algorithm Images"""

import numpy
from rio_tiler.models import ImageData

from titiler.core.algorithm.base import BaseAlgorithm


class ToGrayScale(BaseAlgorithm):
    """Transform a RGB Image to Grayscale."""

    title: str = "Transform a RGB Image to Grayscale"
    description: str = "Transform a RGB Image to Grayscale using the ITU-R 601-2 luma."

    # metadata
    output_nbands: int = 1

    def __call__(self, img: ImageData) -> ImageData:
        """RGB to L."""
        if img.count < 3:
            raise ValueError(
                f"Cannot apply `grayscale` algorithm on image with {img.count} bands."
            )

        arr = (
            img.array[0] * 299 / 1000
            + img.array[1] * 587 / 1000
            + img.array[2] * 114 / 1000
        )
        return ImageData(
            arr.astype(img.array.dtype),
            assets=img.assets,
            crs=img.crs,
            band_descriptions=["grayscale"],
            bounds=img.bounds,
            cutline_mask=img.cutline_mask,
        )


class ToBitonal(BaseAlgorithm):
    """Transform an Image to Bitonal."""

    title: str = "Transform an Image to Bitonal"
    description: str = "All values larger than 127 are set to 255 (white), all other values to 0 (black)."

    # metadata
    output_nbands: int = 1
    output_dtype: str = "uint8"

    def __call__(self, img: ImageData) -> ImageData:
        """Image to Bitonal"""
        if img.count == 3:
            # Convert to Grayscale
            arr = (
                img.array[0] * 299 / 1000
                + img.array[1] * 587 / 1000
                + img.array[2] * 114 / 1000
            )
        elif img.count == 1:
            arr = img.array
        else:
            raise ValueError(
                f"Cannot apply `bitonal` algorithm on image with {img.count} bands."
            )

        return ImageData(
            numpy.ma.where(arr > 127, 255, 0).astype("uint8"),
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
            band_descriptions=["bitonal"],
            cutline_mask=img.cutline_mask,
        )



"""titiler.core.algorithm."""

import json
from copy import copy
from typing import Annotated, Literal

import attr
from fastapi import HTTPException, Query
from pydantic import BeforeValidator, ValidationError

from titiler.core.algorithm.base import (  # noqa
    AlgorithmMetadata,
    AlgorithmtList,
    BaseAlgorithm,
)
from titiler.core.algorithm.dem import Contours, HillShade, Slope, TerrainRGB, Terrarium
from titiler.core.algorithm.image import ToBitonal, ToGrayScale
from titiler.core.algorithm.index import NormalizedIndex
from titiler.core.algorithm.math import _Max, _Mean, _Median, _Min, _Std, _Sum, _Var
from titiler.core.algorithm.ops import CastToInt, Ceil, Floor
from titiler.core.validation import validate_json

default_algorithms: dict[str, type[BaseAlgorithm]] = {
    "hillshade": HillShade,
    "slope": Slope,
    "contours": Contours,
    "normalizedIndex": NormalizedIndex,
    "terrarium": Terrarium,
    "terrainrgb": TerrainRGB,
    "cast": CastToInt,
    "ceil": Ceil,
    "floor": Floor,
    "min": _Min,
    "max": _Max,
    "median": _Median,
    "mean": _Mean,
    "std": _Std,
    "var": _Var,
    "sum": _Sum,
    "grayscale": ToGrayScale,
    "bitonal": ToBitonal,
}


@attr.s(frozen=True)
class Algorithms:
    """Algorithms."""

    data: dict[str, type[BaseAlgorithm]] = attr.ib(factory=dict)

    def get(self, name: str) -> type[BaseAlgorithm]:
        """Fetch a TMS."""
        if name not in self.data:
            raise KeyError(f"Invalid name: {name}")

        return self.data[name]

    def list(self) -> list[str]:
        """List registered Algorithm."""
        return list(self.data.keys())

    def register(
        self,
        algorithms: dict[str, BaseAlgorithm],
        overwrite: bool = False,
    ) -> "Algorithms":
        """Register Algorithm(s)."""
        for name, _algo in algorithms.items():
            if name in self.data and not overwrite:
                raise Exception(f"{name} is already a registered. Use overwrite=True.")

        return Algorithms({**self.data, **algorithms})  # type: ignore [dict-item]

    @property
    def dependency(self):
        """FastAPI PostProcess dependency."""

        def post_process(
            algorithm: Annotated[
                Literal[tuple(self.data.keys())],
                Query(description="Algorithm name"),
            ] = None,
            algorithm_params: Annotated[
                str | None,
                BeforeValidator(validate_json),
                Query(description="Algorithm parameter"),
            ] = None,
        ) -> BaseAlgorithm | None:
            """Data Post-Processing options."""
            kwargs = json.loads(algorithm_params) if algorithm_params else {}
            if algorithm:
                try:
                    return self.get(algorithm)(**kwargs)

                except ValidationError as e:
                    raise HTTPException(status_code=400, detail=str(e)) from e

            return None

        return post_process


algorithms = Algorithms(copy(default_algorithms))  # noqa
